#Coarse mesh [com.nx=8, com.ny=4] for DIII-D MHD equilibrium
#Uses diffusive neutrals, so five variables [bbb.ni,bbb.upi,Te,Ti,bbb.ng]
#
import sys,os
from uedge import *
# Set the com.geometry
bbb.mhdgeo = 1 		#use MHD equilibrium
#com.rm -flx.f aeqdsk neqdsk		#change names of MHD eqil. files 

flx.psi0min1 = 0.98		#normalized flux on core bndry
flx.psi0min2 = 0.98		#normalized flux on pf bndry
flx.psi0sep = 1.00001	#normalized flux at separatrix
flx.psi0max = 1.07		#normalized flux on outer wall bndry
bbb.ngrid = 1		#number of meshes [always set to 1]
com.nxleg[0,0] = 4		#pol. mesh pts from inner plate to flx.x-point
com.nxcore[0,0] = 4		#pol. mesh pts from flx.x-point to top on inside
com.nxcore[0,1] = 4		#pol. mesh pts from top to flx.x-point on outside
com.nxleg[0,1] = 4		#pol. mesh pts from flx.x-point to outer plate
com.nysol[0] = 6		#rad. mesh pts in SOL
com.nycore[0] =2		#rad. mesh pts in core

# Finite-difference algorithms [upwind, central diff, etc.]
bbb.methn = 33		#ion continuty eqn
bbb.methu = 33		#ion parallel momentum eqn
bbb.methe = 33		#electron energy eqn
bbb.methi = 33		#ion energy eqn
bbb.methg = 33		#neutral gas continuity eqn

# Boundary conditions
bbb.ncore[0] = 2.e19	#hydrogen ion density on core
bbb.iflcore = 0		#flag =0, fixed Te,i =1, fixed power on core
bbb.tcoree = 100.		#core Te 
bbb.tcorei = 100.		#core Ti
bbb.tedge = 2.		#fixed wall,pf Te,i if istewcon=1, etc
bbb.istepfc = 3		#=3 sets scale length bbb.lyte
bbb.lyte = 0.03		#radial scale-length for Te on PF boundary
bbb.recycp[0] = 0.98		#hydrogen recycling grd.coeff at plates
bbb.recycw[0] = 0.9		#wall recycling if bbb.matwso,i=1
bbb.matwso[0] = 1		#recycle on main-chamber wall
bbb.isnwcono = 1		#if=1, set bbb.ni[,,com.ny+1]=bbb.nwallo
bbb.isnwconi = 1		#if=1, set PF wall bbb.ni=bbb.nwalli
bbb.allocate()		#bbb.allocate() space of bbb.nwallo,i
bbb.nwallo = 1.e18
bbb.nwalli = 1.e18

# Transport coefficients
bbb.difni[0] = 1.		#D for radial hydrogen diffusion
bbb.kye = 1.		#chi_e for radial elec energy diffusion
bbb.kyi = 1.		#chi_i for radial ion energy diffusion
bbb.travis[0] = 1.		#eta_a for radial ion momentum diffusion

# Flux limits
bbb.flalfe = 0.21		#electron parallel thermal conduct. grd.coeff
bbb.flalfi = 0.21		#ion parallel thermal conduct. grd.coeff
bbb.flalfv = 1.		#ion parallel viscosity grd.coeff
bbb.flalfgx = 1.		#neut. density poloidal diffusion
bbb.flalfgy = 1.		#neut. density radial diffusion
bbb.flalfvgx = 1.		#neut. momentum poloidal viscosity
bbb.flalfvgy = 1.		#neut. momentum radial viscosity
bbb.flalftgx = 1.		#neut. particle poloidal heat diffusion 
bbb.flalftgy = 1.		#neut. particle radial heat diffusion

# Solver package
bbb.svrpkg = "nksol"	#Newton solver using Krylov method
bbb.premeth = "ilut"	#Solution method for precond. Jacobian matrix
bbb.mfnksol = 3

# Parallel neutral momentum equation
bbb.ineudif = 2
bbb.isupgon[0]=1
if bbb.isupgon[0] == 1:
   bbb.isngon[0]=0
   com.ngsp=1
   com.nhsp=2
   bbb.ziin[com.nhsp-1]=0
   bbb.travis[1] = 0.	#shouldn't be used for neutrals - to be sure
   # The following are probably default, set them anyway to be sure
   bbb.cngmom=0
   bbb.cmwall=0
   bbb.cngtgx=0
   bbb.cngtgy=0
   bbb.kxn=0
   bbb.kyn=0
##	bbb.ingb = 0

# Currents and potenial parameters
bbb.isphion = 1
bbb.b0 = 1.			# =1 for normal direction of B-field
bbb.rsigpl=1.e-8		#anomalous cross-field conductivity
bbb.cfjhf=1.		#turn-on heat flow from current [bbb.fqp]
bbb.cfjve = 1.		#makes bbb.vex = vix - bbb.cfjve*bbb.fqx
bbb.jhswitch=1		#Joule Heating switch
bbb.cfjpy = 0.		#diamag. cur. in flx.y-direction
bbb.cfjp2 = 0.		#diamag. cur. in 2-direction

bbb.newbcl=1
bbb.newbcr=1	#Sheath BC [bee,i] from current equation
bbb.isfdiax =1.		#Factor to turn on diamag. contrib. to sheath

bbb.cfyef = 1.0		#ExB drift in flx.y-dir.
bbb.cf2ef = 1.0		#ExB drift in 2-dir.
bbb.cfydd = 0.		#Diamag. drift in flx.y-dir. [always=0]
bbb.cf2dd = 0.		#Diamag. drift in 2-dir. [always=0]
bbb.cfrd = 0.		#Resistive drift in flx.y and 2 dirs.
bbb.cfbgt = 0.		#Diamag. energy drift [always=0]
bbb.cfybf = 1.              #turns on bbb.vycb - radial grad_B drift
bbb.cf2bf = 1.              #turns on bbb.v2cb - perp grad_B drift [nearly pol]
bbb.cfqybf = 1.             #turns on bbb.vycb contrib to radial current
bbb.cfq2bf = 1.             #turns on bbb.v2cb contrib to perp["2"] current
bbb.isnewpot = 1
bbb.rnewpot = 1.		
bbb.iphibcc = 0		#set bbb.phi[,1] uniform poloidally		

# Restart from bbb.a pfb savefile
bbb.restart = 1		#Begin from savefile, not estimated profiles
bbb.allocate()		#bbb.allocate() space for savevariables



# Import all required libraries

import numpy
import pylab
import string

#
# This class will generate a lineout of the given data across the
# midplane starting at the magnetic axis and going all the way to the
# edge.  Eventually it should be capable of generating a lineout along
# any given ray, but currently only midplane is implemented
#
class UeGenLineout:

#
# Constructor
#
    def __init__(self):
        pass

#
# Use this to set the core region data and grids
#
# @param xx the XX array of the mapped grid
# @param yy the YY array of the mapped grid
# @param da the data at each x,y point described by xx and yy
#
    def setCoreData(self, xx, yy, da):
        self.eqXX = xx
        self.eqYY = yy
        self.eqDA = da

#
# Use this to set the interior edge region data and grids
#
# @param xx the XX array of the mapped grid centers
# @param yy the YY array of the mapped grid centers
# @param da the data at each x,y point described by xx and yy
#
    def setInteriorEdgeData(self, xx, yy, da):
        self.coXX = xx
        self.coYY = yy
        self.coDA = da

#
# Use this to set the scrape-off layer region data and grids
#
# @param xx the XX array of the mapped grid
# @param yy the YY array of the mapped grid
# @param da the data at each x,y point described by xx and yy
#
    def setSOLEdgeData(self, xx, yy, da):
        self.scXX = xx
        self.scYY = yy
        self.scDA = da

#
# Use this to get a lineout of the data input.  Call all the set
# functions before calling this
#
# @param fname file from which to read processed uedge grid points
#
# @return a tuple of numpy arrays representing the lineout data
#     1: The array of X coordinates of the data
#     2: The array of data representing the lineout
#
    def getLineout(self, fname):
        eqRetX = self.eqXX[0,:]
        eqRetD = self.eqDA[0,:]

        fh = open(fname, 'r')

        lns = fh.readlines()

        da = numpy.zeros((lns.__len__(), 5), numpy.float)
        for i in range(0, lns.__len__()):
            ln = lns[i].split()
            for j in range(0, ln.__len__()):
                da[i, j] = string.atof( ln[j] )

        n = 0
        for i in range(0, self.coXX.shape[1]):
            arrX = self.coXX[:,i]
            arrY = self.coYY[:,i]
            arrD = self.coDA[:,i]

            q = 0;
            for j in range (1, arrY.shape[0]):
                if ( ( arrX[j-1] > eqRetX[0] ) and
                     ( ( arrY[j-1] < da[n, 1] and
                         arrY[j]   > da[n, 1] ) or
                       ( arrY[j]   < da[n, 1] and
                         arrY[j-1] > da[n, 1] ) ) ):
                    q = j
                    break

            rat = (da[n, 1] - arrY[q-1]) / (arrY[q] - arrY[q-1])
            da[n, 4] = rat * arrD[q] + (1-rat)*arrD[q-1]

            n = n+1

        for i in range(0, self.scXX.shape[1]):
            arrX = self.scXX[:,i]
            arrY = self.scYY[:,i]
            arrD = self.scDA[:,i]

            q = 0;
            for j in range (1, arrY.shape[0]):
                if ( ( arrX[j-1] > eqRetX[0] ) and
                     ( ( arrY[j-1] < da[n, 1] and
                         arrY[j]   > da[n, 1] ) or
                       ( arrY[j]   < da[n, 1] and
                         arrY[j-1] > da[n, 1] ) ) ):
                    q = j
                    break

            rat = (da[n, 1] - arrY[q-1]) / (arrY[q] - arrY[q-1])
            da[n, 4] = rat * arrD[q] + (1-rat)*arrD[q-1]

            n = n+1

        retX = numpy.zeros( (eqRetX.shape[0] + da.shape[0] - 1), numpy.float )
        retD = numpy.zeros( (eqRetX.shape[0] + da.shape[0] - 1), numpy.float )

        retX[ :eqRetX.shape[0]-1 ] = eqRetX[:-1]
        retX[ eqRetX.shape[0]-1: ] = da[:,0]

        retD[ :eqRetX.shape[0]-1 ] = eqRetD[:-1]
        retD[ eqRetX.shape[0]-1: ] = da[:,4]
        
        return retX, retD



#!/usr/bin/env python

import convertvsh5

import sys
import optparse

cpsr = optparse.OptionParser()
cpsr.add_option( '-i', '--input-file', dest='file', action='store',
                 help='Specify the file to annotate' )
cpsr.add_option( '-o', '--output-file', dest='ofile', action='store',
                 help='Specify the file to write to' )

opts, args = cpsr.parse_args(sys.argv)

cvtr = convertvsh5.UeConvertVsH5()
cvtr.readData(opts.file)
cvtr.createUnstructGrid()
cvtr.annotateFile(opts.ofile)
cvtr.annotateStructured(opts.ofile)



# This file runs a time-dependent case using dtreal.  First, obtain a converged
# solution for a (usually small) dtreal; xuedge must report iterm=1 at the end.
# Then adjust control parameters in rdinitdt; read this file, which reads rdinitdt.
# If a mistake is made, to restart this file without a Jacobian evaluation,
# be sure to reset iterm=1 (=> last step was successful)

# IMPORT UEDGE (assuming starting from ipython before any imports)
from uedge import *
        
# IMPORT HDF5 routines for saving solutions below
from uedge.hdf5 import *

# INITIALIZE PARAMS -- SHOULD BE DONE IN MASTER SCRIPT OR TERMINAL SESSION
# BEFORE INVOKING THIS SCRIPT
execfile("rdinitdt.py")
no = 0;yes = 1
echo = no

# Set precisions of floating point output
###import print_options
###print_options.set_float_precision(4)

# Check if successful time-step exists (bbb.iterm=1)
if (bbb.iterm == 1):
    print "Initial successful time-step exists"
    bbb.dtreal = bbb.dtreal*bbb.mult_dt #compensates dtreal divided by mult_dt below
else:
    print "*---------------------------------------------------------*"
    print "Need to take initial step with Jacobian; trying to do here"
    print "*---------------------------------------------------------*"
    bbb.icntnunk = 0
    bbb.exmain()
    bbb.dtreal = bbb.dtreal*bbb.mult_dt #compensates dtreal divided by mult_dt below

if (bbb.iterm <> 1):
    print "*--------------------------------------------------------------*"
    print "Error: converge an initial time-step first; then retry rdcontdt"
    print "*--------------------------------------------------------------*"
    exit()
 
nx=com.nx;ny=com.ny;nisp=com.nisp;ngsp=com.ngsp;numvar=bbb.numvar
isteon=bbb.isteon
if (i_stor==0):
   ni_stor = zeros((bbb.n_stor,nx+1+1,ny+1+1,nisp),"d")	# set time storage arrays
   up_stor = zeros((bbb.n_stor,nx+1+1,ny+1+1,nisp),"d")
   te_stor = zeros((bbb.n_stor,nx+1+1,ny+1+1),"d")
   ti_stor = zeros((bbb.n_stor,nx+1+1,ny+1+1),"d")
   ng_stor = zeros((bbb.n_stor,nx+1+1,ny+1+1,ngsp),"d")
   phi_stor = zeros((bbb.n_stor,nx+1+1,ny+1+1),"d")
   tim_stor = zeros((bbb.n_stor),"d")
   dtreal_stor = zeros((bbb.n_stor),"d")
   nfe_stor = zeros((bbb.n_stor),"l")
   dt_stor = (bbb.tstor_e - bbb.tstor_s)/(bbb.n_stor - 1)

i_stor = max(i_stor,1)			# set counter for storage arrays
bbb.dt_tot = max(bbb.dt_tot,0.)
nfe_tot = max(nfe_tot,0)
deldt_0 = bbb.deldt
isdtsf_sav = bbb.isdtsfscal

if (bbb.ipt==1 and bbb.isteon==1): 	# set ipt to te(nx,iysptrx+1) if no user value
   ipt = bbb.idxte[nx-1,com.iysptrx]  #note: ipt is local, bbb.ipt global

bbb.irev = -1         # forces second branch of irev in ii1 loop below
if (bbb.iterm == 1):  # successful initial run with dtreal
   bbb.dtreal = bbb.dtreal/bbb.mult_dt     # gives same dtreal after irev loop
else:                 # unsuccessful initial run; reduce dtreal
   bbb.dtreal = bbb.dtreal/(3*bbb.mult_dt) # causes dt=dt/mult_dt after irev loop
   
if (bbb.initjac == 0): bbb.newgeo=0
dtreal_sav = bbb.dtreal
bbb.itermx = bbb.itermxrdc
bbb.dtreal = bbb.dtreal/bbb.mult_dt	#adjust for mult. to follow; mult_dt in rdinitdt
bbb.dtphi = bbb.rdtphidtr*bbb.dtreal
neq=bbb.neq
svrpkg=bbb.svrpkg.tostring().strip()
#
bbb.ylodt = bbb.yl
bbb.pandf1 (-1, -1, 0, bbb.neq, 1., bbb.yl, bbb.yldot)
fnrm_old = sqrt(sum((bbb.yldot[0:neq]*bbb.sfscal[0:neq])**2))
if (bbb.initjac == 1): fnrm_old=1.e20
print "initial fnrm =",fnrm_old

for ii1 in range( 1, bbb.ii1max+1):
   if (bbb.ismfnkauto==1): bbb.mfnksol = 3
   # adjust the time-step
   if (bbb.irev == 0):
      # Only used after a dt reduc. success. completes loop ii2 for fixed dt
      bbb.dtreal = min(3*bbb.dtreal,bbb.t_stop)	#first move forward after reduction
      bbb.dtphi = bbb.rdtphidtr*bbb.dtreal
      if (bbb.ismfnkauto==1 and bbb.dtreal > bbb.dtmfnk3): bbb.mfnksol = -3
      bbb.deldt =  3*bbb.deldt
   else:
      # either increase or decrease dtreal; depends on mult_dt
      bbb.dtreal = min(bbb.mult_dt*bbb.dtreal,bbb.t_stop)
      bbb.dtphi = bbb.rdtphidtr*bbb.dtreal
      if (bbb.ismfnkauto==1 and bbb.dtreal > bbb.dtmfnk3): bbb.mfnksol = -3
      bbb.deldt =  bbb.mult_dt*bbb.deldt
      
   bbb.dtreal = min(bbb.dtreal,bbb.dt_max)
   bbb.dtphi = bbb.rdtphidtr*bbb.dtreal
   if (bbb.ismfnkauto==1 and bbb.dtreal > bbb.dtmfnk3): bbb.mfnksol = -3
   bbb.deldt = min(bbb.deldt,deldt_0)
   bbb.deldt = max(bbb.deldt,bbb.deldt_min)
   nsteps_nk=1
   print '--------------------------------------------------------------------'
   print '--------------------------------------------------------------------'
   print ' '
   print '*** Number time-step changes = ',ii1,' New time-step = ', bbb.dtreal
   print '--------------------------------------------------------------------'

   bbb.itermx = bbb.itermxrdc
   if (ii1>1  or  bbb.initjac==1):	# first time calc Jac if initjac=1
      if (bbb.irev == 1):      # decrease in bbb.dtreal
         if (bbb.numrev < bbb.numrevjmax and \
            bbb.numrfcum < bbb.numrevjmax+bbb.numfwdjmax): #dont recom bbb.jac
            bbb.icntnunk = 1	
            bbb.numrfcum = bbb.numrfcum + 1
         else:                          # force bbb.jac calc, reset numrev
            bbb.icntnunk = 0
            bbb.numrev = -1		      # yields api.zero in next statement
            bbb.numrfcum = 0
         bbb.numrev = bbb.numrev + 1
         bbb.numfwd = 0
      else:  # increase in bbb.dtreal
         if (bbb.numfwd < bbb.numfwdjmax and \
            bbb.numrfcum < bbb.numrevjmax+bbb.numfwdjmax): 	#dont recomp bbb.jac
            bbb.icntnunk = 1
            bbb.numrfcum = bbb.numrfcum + 1
         else:
            bbb.icntnunk = 0			#recompute jacobian for increase dt
            bbb.numfwd = -1
            bbb.numrfcum = 0
         bbb.numfwd = bbb.numfwd + 1
         bbb.numrev = 0			#bbb.restart counter for dt reversals
      bbb.isdtsfscal = isdtsf_sav
      bbb.ftol = min(bbb.ftol_dt, 0.01*fnrm_old)
      exmain() # take a single step at the present bbb.dtreal
      if (bbb.iterm == 1):
         bbb.dt_tot = bbb.dt_tot + bbb.dtreal
         nfe_tot = nfe_tot + bbb.nfe[0,0]
         bbb.ylodt = bbb.yl
         bbb.pandf1 (-1, -1, 0, bbb.neq, 1., bbb.yl, bbb.yldot)
         fnrm_old = sqrt(sum((bbb.yldot[0:neq-1]*bbb.sfscal[0:neq-1])**2))
         if (bbb.dt_tot>=0.9999999*bbb.t_stop  or  fnrm_old<bbb.ftol_min):
            print ' '
            print '*****************************************************'
            print '**  SUCCESS: frnm < bbb.ftol; or dt_tot >= t_stop  **'
            print '*****************************************************'
            break

   bbb.icntnunk = 1
   bbb.isdtsfscal = 0
   for ii2 in range( 1, bbb.ii2max+1): #take ii2max steps at the present time-step
      if (bbb.iterm == 1):
         bbb.itermx = bbb.itermxrdc
         bbb.ftol = min(bbb.ftol_dt, 0.01*fnrm_old)
         bbb.exmain()
         if (bbb.iterm == 1):
            bbb.ylodt = bbb.yl
            bbb.pandf1 (-1, -1, 0, bbb.neq, 1., bbb.yl, bbb.yldot)
            fnrm_old = sqrt(sum((bbb.yldot[0:neq-1]*bbb.sfscal[0:neq-1])**2))
            print "Total time = ",bbb.dt_tot,"; Timestep = ",bbb.dtreal
            print "variable index ipt = ",ipt, " bbb.yl[ipt] = ",bbb.yl[ipt]
            dtreal_sav = bbb.dtreal
            bbb.dt_tot = bbb.dt_tot + bbb.dtreal
            nfe_tot = nfe_tot + bbb.nfe[0,0]
            hdf5_save(savefn)
            if (bbb.dt_tot>=0.999999999999*bbb.t_stop  or  fnrm_old<bbb.ftol_min):
               print ' '
               print '*****************************************************'
               print '**  SUCCESS: frnm < bbb.ftol; or dt_tot >= t_stop  **'
               print '*****************************************************'
               break
            print " "
##       Store variables if a storage time has been crossed
            if (bbb.dt_tot >= dt_stor*i_stor and i_stor<=bbb.n_stor):
               i_stor1 = i_stor-1
               ni_stor[i_stor1,:,:,:] = ni
               up_stor[i_stor1,:,:,:] = up
               te_stor[i_stor1,:,:] = te
               ti_stor1[i_stor1,:,:] = ti
               ng_stor[i_stor1,:,:,:] = ng
               phi_stor1[i_stor1,:,:] = phi
               tim_stor[i_stor1] = bbb.dt_tot
               nfe_stor[i_stor1] = nfe_tot
               dtreal_stor[i_stor1] = bbb.dtreal
               i_stor = i_stor + 1
   ##          End of storage section
      
   if (bbb.dt_tot>=bbb.t_stop  or  fnrm_old<bbb.ftol_min): break   # need for both loops
   bbb.irev = bbb.irev-1
   if (bbb.iterm != 1):	#print bad eqn, cut dtreal by 3, set irev flag
            ####### a copy of idtroub script ########################
      oldecho=echo
      echo=no
      # integer ii
      # real8 ydmax 
      scalfac = bbb.sfscal
      if (svrpkg != "nksol"): scalfac = 1/(bbb.yl + 1.e-30)  # for time-dep calc.
      ydmax = 0.999999999*max(abs(bbb.yldot*scalfac))
      itrouble = 0
      for ii in range(neq):
         if (abs(bbb.yldot[ii]*scalfac[ii]) > ydmax):
            itrouble=ii
            print "** Fortran index of trouble making equation is:"
            print itrouble+1
            break
      print "** Number of variables is:"
      print "numvar = ", numvar
      print " "
      iv_t = (itrouble).__mod__(numvar) + 1
      print "** Troublemaker equation is:"
      print "iv_t = ",iv_t
      print " "
      print "** Troublemaker cell (ix,iy) is:"
      print bbb.igyl[itrouble,]
      print " "
      print "** Timestep for troublemaker equation:"
      print bbb.dtuse[itrouble]
      print " "
      print "** yl for troublemaker equation:"
      print bbb.yl[itrouble]
      print " "
      echo=oldecho
      ######## end of idtroub script ##############################

      if (bbb.dtreal < bbb.dt_kill):
          print ' '
          print '*************************************'
          print '**  FAILURE: time-step < dt_kill   **'
          print '*************************************'
          break
      bbb.irev = 1
      print '*** Converg. fails for bbb.dtreal; reduce time-step by 3, try again'
      print '----------------------------------------------------------------- '
      bbb.dtreal = bbb.dtreal/(3*bbb.mult_dt)
      bbb.dtphi = bbb.rdtphidtr*bbb.dtreal
      if (bbb.ismfnkauto==1 and bbb.dtreal > bbb.dtmfnk3): bbb.mfnksol = -3
      bbb.deldt =  bbb.deldt/(3*bbb.mult_dt) 
      bbb.iterm = 1
echo = yes
